# -*- coding: utf-8 -*-
"""
Apply trained best-performing (BP) SNGLT model. The SNGLT model was trained during HP tuning. The overall best performer for each loss function was chosen by training MULTT on the five best configurations for SNGLT and choosing the configs that had lowest RMSE overall.

@author: Kathy_Breen
"""

from keras.layers import Input, Dense, Concatenate, Dropout, BatchNormalization, CuDNNLSTM
from keras.models import Model
from keras.callbacks import EarlyStopping, ReduceLROnPlateau, CSVLogger, ModelCheckpoint
from keras.initializers import Ones
from keras.optimizers import Adagrad, SGD, RMSprop, Adadelta, Adamax, Adam, Nadam
from sklearn.metrics import mean_squared_error
from keras.models import load_model
from keras.utils import CustomObjectScope
import tensorflow as tf
from keras import backend as K
import h5py
import numpy as np
import pandas as pd
import pickle as pkl
from matplotlib import rcParams  # next 3 lines set font family for plotting
rcParams['font.family'] = 'serif'
rcParams['font.sans-serif'] = ['Times New Roman']
import matplotlib.pyplot as plt
# increase font size
plt.rcParams.update({'font.size': 22})
import os
import datetime

from scipy import stats

#%% SETTINGS FOR REPRODUCIBLE RESULTS DURING DEVELOPMENT
'''
#import numpy as np
import tensorflow as tf
import random as rn

# The below is necessary in Python 3.2.3 onwards to
# have reproducible behavior for certain hash-based operations.
# See these references for further details:
# https://docs.python.org/3.4/using/cmdline.html#envvar-PYTHONHASHSEED
# https://github.com/keras-team/keras/issues/2280#issuecomment-306959926

#import os
os.environ['PYTHONHASHSEED'] = '0'

# The below is necessary for starting Numpy generated random numbers
# in a well-defined initial state.

np.random.seed(42)

# The below is necessary for starting core Python generated random numbers
# in a well-defined state.

rn.seed(12345)

# Force TensorFlow to use single thread.
# Multiple threads are a potential source of
# non-reproducible results.
# For further details, see: https://stackoverflow.com/questions/42022950/which-seeds-have-to-be-set-where-to-realize-100-reproducibility-of-training-res

session_conf = tf.compat.v1.ConfigProto(intra_op_parallelism_threads=1, inter_op_parallelism_threads=1)

from keras import backend as K

# The below tf.set_random_seed() will make random number generation
# in the TensorFlow backend have a well-defined initial state.
# For further details, see: https://www.tensorflow.org/api_docs/python/tf/set_random_seed

#tf.set_random_seed(1234)
tf.compat.v1.set_random_seed(1234)

sess = tf.compat.v1.Session(graph=tf.compat.v1.get_default_graph(), config=session_conf)
K.set_session(sess)
'''
#%%
# define custom loss function

def sameSign(x,y):
    # returns 0 if x and y are of opposite sign, and 1 if the signs are the same
    xsign = tf.sign(x)
    ysign = tf.sign(y)
    
    return tf.to_float(tf.equal(xsign,ysign))
    

def pblf(ET_DATA,ET_SHIFT):
    
    # ET_DATA is an array the same size as y_true with daily ET values temporally sync all samples     
    # ET_DATA is an array the same size as y_true with daily ET values temporally sync all samples in Y
    # Y_SHIFT is an array the same size as y_true where y_true[i,:] == Y_SHIFT[i-1,:]
    # ET_SHIFT is an array the same size as ET_DATA where ET_DATA[i,:] == ET_SHIFT[i-1,:]
    
    def loss(y_true,y_pred):
    
        Y_SHIFT = tf.roll(y_pred,1,axis=0)

        y_pred_diff = tf.subtract(y_pred,Y_SHIFT)
        et_diff = tf.subtract(ET_DATA,ET_SHIFT)
        
        # test if soil moisture decreased when et increased and v.v. - penalize loss function if physical relationship does not hold, i.e. the signs are the same
        delta_sign = sameSign(y_pred_diff,et_diff)
        return K.mean(K.square(y_pred-y_true) + 0.1*delta_sign*K.sqrt(K.square(et_diff) + K.square(y_pred_diff)),axis=-1)
    
    return loss

#%%
    
# read and plot output from best performing models
bp_basenum = '47'  # best performer with MSE
#bp_basenum = '23'  # best performer with PBLF (also overall)
loss_func = 'MSE'
#loss_func = 'PBLF'
flag = ''
#flag = '_noflag'
print('DEFINE WORKING DIRECTORY PATH AS THE VARIABLE "MAIN"')
main = ''
os.chdir(main)
print('STORE FILES WITH "BP" IN THE FILENAME IN A FOLDER NAMED "BEST PERFORMERS"')
bp_main = os.path.join(main,'BestPerformers')
    
os.chdir(bp_main)
best_model = 'BP_model_' + bp_basenum + flag + '.h5'
BPconfig = pd.read_csv('BPout.csv')
idxx = BPconfig[BPconfig['Basename']=='BP_model_'+bp_basenum].index[0]
        
# read in model and hist dict
y_et = Input(shape=(1096,), name='y_et')
et_shift = Input(shape=(1096,), name='et_shift')
    
print(loss_func+flag)
if loss_func == 'PBLF':
    with CustomObjectScope({'loss': pblf(ET_DATA=y_et,ET_SHIFT=et_shift)}):
        model = load_model(best_model)
else:
    model = load_model(best_model)

histdict = pkl.load(open('BP_histdict_' + bp_basenum + flag + '.pkl',"rb"))
os.chdir(main)

#%%
Y_rmse_out = {}
timeseries_out = {}
stats_out = {}

# for sites where the trained model will be tested, define site id
test_ids = ['ELKR','YAZO','WASH','ESCL']
for ip, test_site_id in enumerate(test_ids):
    print (test_site_id)
    
    Y_rmse_out[test_site_id] = {}
    
    if test_site_id == 'ELKR':
            
        with h5py.File('Test0'+flag+'.hdf5','r') as f:
            XD = np.array(f['XD_test0'])
            XS = np.array(f['XS_test0'])
            Ysmap = np.array(f['Y_test0'])
            Yet = np.array(f['Yet_test0'])
            Yet_shift = np.array(f['Yet_shift_test0'])
        
    else:
        test_main = os.path.join(main,test_site_id)
        os.chdir(test_main)
                
        with h5py.File('Test0_'+test_site_id+flag+'.hdf5','r') as f:
            XD = np.array(f['XD_test0'])
            XS = np.array(f['XS_test0'])
            Ysmap = np.array(f['Y_test0'])
            Yet = np.array(f['Yet_test0'])
            Yet_shift = np.array(f['Yet_shift_test0'])
    
    # the model is trained on 1096 timesteps - 01/01/2016-12/31/2018. the input arrays have data from 01/01/16-12/31/19 the output arrays all have data from 3/31/15-12/31/19 - limit the data I/O to the time period for training. To predict "future" SM, use the last 1096 timesteps of each array [-1096:] 
    XD = XD[:,:1096,:]   
    Ysmap = Ysmap[:,276:276+1096]
    Yet = Yet[:,276:276+1096]
    Yet_shift = Yet_shift[:,276:276+1096]
    
    # load previously trained model
    predict = model.predict([XD,XS],batch_size=BPconfig.loc[idxx,'batch_size'],verbose=1)
    
    Yhat_smap_final = predict[0]  # final output predictions to compare with Ysmap
    Yhat_smap_aux = predict[1]  # auxilliary output predictions to compare with Ysmap
    
    # perform predictions for each of the test sites and write the results to file
    for test_site_idx in range(XD.shape[0]):
        Y_rmse_final = np.sqrt(mean_squared_error(Ysmap[test_site_idx],Yhat_smap_final[test_site_idx]))
        Y_rmse_aux = np.sqrt(mean_squared_error(Ysmap[test_site_idx],Yhat_smap_aux[test_site_idx]))
        
        # calculate moving average/std dev fr SM same way it was done for input data
        window_size = 90  # calc moving avg using a window of 10 days
        stride = int(window_size*.5)  # overlap windows by 50%
        ts = Ysmap[test_site_idx,:]
        ts_mu = np.zeros(ts.shape)
        ts_sigma = np.zeros(ts.shape) 
        window_beg_idx = list(range(0, len(ts), stride))
        ts_ma = [np.mean(ts[i:i+window_size]) for i in window_beg_idx if i+window_size <= len(ts)]
        ts_std = [np.std(ts[i:i+window_size]) for i in window_beg_idx if i+window_size <= len(ts)]
        for i,idx in enumerate(window_beg_idx):
        #            print(i,i+window_size)
            if idx+window_size <= len(ts):
                ts_mu[idx:idx+window_size] = ts_ma[i]
                ts_sigma[idx:idx+window_size] = ts_std[i]
            else:
                ts_mu[idx:len(ts)] = ts_ma[len(ts_ma)-1]
                ts_sigma[idx:len(ts)] = ts_std[len(ts_std)-1]
                
        # predict RMSE for prediction of extreme values and "normal" values separately
        plus_sigma_idx = np.where(Ysmap[test_site_idx,:] >= ts_mu + ts_sigma)
        minus_sigma_idx = np.where(Ysmap[test_site_idx,:] <= ts_mu - ts_sigma)
        normal_idx = np.where((Ysmap[test_site_idx,:] > ts_mu - ts_sigma) & (Ysmap[test_site_idx,:] < ts_mu + ts_sigma))
        np.where(Ysmap[test_site_idx,:] < ts_mu + ts_sigma)[0]
        a = np.concatenate([Ysmap[test_site_idx,plus_sigma_idx],Ysmap[test_site_idx,minus_sigma_idx]],axis=1)
        b = np.concatenate([Yhat_smap_final[test_site_idx,plus_sigma_idx],Yhat_smap_final[test_site_idx,minus_sigma_idx]],axis=1)
        Y_rmse_extreme_final = np.sqrt(mean_squared_error(a,b))
        b = np.concatenate([Yhat_smap_aux[test_site_idx,plus_sigma_idx],Yhat_smap_aux[test_site_idx,minus_sigma_idx]],axis=1)
        Y_rmse_extreme_aux = np.sqrt(mean_squared_error(a,b))
        a = Ysmap[test_site_idx,normal_idx]
        b = Yhat_smap_final[test_site_idx,normal_idx]
        Y_rmse_normal_final = np.sqrt(mean_squared_error(a,b))
        b = Yhat_smap_aux[test_site_idx,normal_idx]
        Y_rmse_normal_aux = np.sqrt(mean_squared_error(a,b))
    
        Y_rmse_out[test_site_id][test_site_idx] = {
                                   'Y_rmse_aux': Y_rmse_aux,
                                   'Y_rmse_final': Y_rmse_final,
                                   'Y_rmse_normal_aux': Y_rmse_normal_aux,
                                   'Y_rmse_normal_final': Y_rmse_normal_final,
                                   'Y_rmse_extreme_aux': Y_rmse_extreme_aux,
                                   'Y_rmse_extreme_final': Y_rmse_extreme_final
                                   }
    
    os.chdir(main)
    # PLOT TIMESERIES OUTPUT FROM AUX LAYER
    base = datetime.datetime.strptime("2016-01-01", "%Y-%m-%d")
    date_lst = [base + datetime.timedelta(days=x) for x in range(1096)]
    xticks = pd.date_range(date_lst[0],date_lst[len(date_lst)-1],freq='QS').tolist()
    xticklabels = []
    for tick in xticks:
        if tick.month == 1:
            label = 'Jan -' + str(tick.year)
            xticklabels.append(label)
        elif tick.month == 7:
            label = 'Jul -' + str(tick.year)
            xticklabels.append(label)
        else:
            xticklabels.append(' ')
    
    # store timeseries and stats
    stats_out[test_site_id] = {}
    
    timeseries_out['DATE'] = []
    for date in date_lst:
        timeseries_out['DATE'].append(date.strftime('%m/%d/%Y'))
    timeseries_out['Y_SMAP_'+test_site_id] = Ysmap.mean(axis=0)
    timeseries_out['Y_AUX_'+test_site_id] = Yhat_smap_aux.mean(axis=0)
    timeseries_out['Y_FINAL_'+test_site_id] = Yhat_smap_final.mean(axis=0)
    
    aux_slope, aux_intercept, aux_r_value, aux_p_value, aux_std_err = stats.linregress(Ysmap.mean(axis=0),Yhat_smap_aux.mean(axis=0))
    final_slope, final_intercept, final_r_value, final_p_value, final_std_err = stats.linregress(Ysmap.mean(axis=0),Yhat_smap_final.mean(axis=0))
    stats_out[test_site_id]['R2_AUX'] = aux_r_value**2
    stats_out[test_site_id]['R2_FINAL'] = final_r_value**2
    
    # PLOT TIMESERIES OUTPUT FROM FINAL LAYER
    fig = plt.figure(figsize=(8,6))
    ax1 = fig.add_subplot(111)
    yhat_aux = ax1.plot(date_lst,Yhat_smap_aux.mean(axis=0),'b-',label=r'$\hat{Y}_{\mathrm{SMAP,aux}}$',linewidth=6)
    yhat_final = ax1.plot(date_lst,Yhat_smap_final.mean(axis=0),'c-',label=r'$\hat{Y}_{\mathrm{SMAP,final}}$',linewidth=6)
    y = ax1.plot(date_lst,Ysmap.mean(axis=0),'k--', label=r'$Y_{\mathrm{SMAP}}$', linewidth=3)
    ax1.set_xlim(left=date_lst[0],right=date_lst[len(date_lst)-1])
    ax1.set_ylim((0.0,0.5))
    ax1.set_xticks(xticks)
    ax1.set_xticklabels(xticklabels,rotation=30)
    ax1.set_xlabel('Date')
    ax1.set_ylabel(r'Soil moisture ($\mathrm{cm}^3 \mathrm{cm}^{-3})$')

    plt.tight_layout()
    plt.savefig('y_yhat_'+test_site_id+'_'+loss_func+flag+'.png', format='png')
    plt.show()
    
    # PLOT Y V YHAT AUX
    fig = plt.figure(num=5, figsize=(8,6))
    ax = fig.add_subplot(111)
    one2one = ax.plot([0.0,0.05,0.1,0.2,0.3,0.4,0.5],[0.0,0.05,0.1,0.2,0.3,0.4,0.5],'k-')
    yvyhat_aux = ax.plot(Ysmap.mean(axis=0),Yhat_smap_aux.mean(axis=0),'b*',label=r'$\hat{Y}_{\mathrm{SMAP,aux}}$',markersize=10)
    yvyhat_final = ax.plot(Ysmap.mean(axis=0),Yhat_smap_final.mean(axis=0),'c*',label=r'$\hat{Y}_{SMAP,final}$',markersize=10)
    ax.set_ylabel(r'$\hat{Y}_{\mathrm{SMAP}}$')
    ax.set_xlabel(r'$Y_{\mathrm{SMAP}}$')
    ax.set_ylim((0.0,0.5))
    ax.set_xlim((0.0,0.5))
    plt.tight_layout()
#    plt.savefig('y_v_yhat_'+test_site_id+'_'+loss_func+flag+'.png')
    plt.show()
    
    plt.close('all')

# write output to file
Y_rmse_out_df = pd.DataFrame.from_dict({(i,j): Y_rmse_out[i][j] 
                           for i in Y_rmse_out.keys() 
                           for j in Y_rmse_out[i].keys()},
                       orient='index')

Y_rmse_out_df.to_csv('RAW_BP_testsites_Yrmse_out'+'_'+loss_func+flag+'.csv', index=True, header=True)

timeseries_df = pd.DataFrame(timeseries_out)
stats_df = pd.DataFrame(stats_out)

timeseries_df.to_csv('SNGLT_TIME_E_'+loss_func+flag+'_TIMESERIES.csv',header=True,index=False)
stats_df.to_csv('SNGLT_TIME_E_'+loss_func+flag+'_STATS.csv',header=True,index=True)

Y_rmse_mean = {}
for site in test_ids:
    tmp = Y_rmse_out_df[[site in s for s in Y_rmse_out_df.index]]
    Y_rmse_mean[site] = {
                       'Y_rmse_aux_mu': tmp['Y_rmse_aux'].mean(),
                       'Y_rmse_aux_sigma': tmp['Y_rmse_aux'].std(),
                       'Y_rmse_final_mu': tmp['Y_rmse_final'].mean(),
                       'Y_rmse_final_sigma': tmp['Y_rmse_final'].std(),
                       'Y_rmse_normal_aux_mu': tmp['Y_rmse_normal_aux'].mean(),
                       'Y_rmse_normal_aux_sigma': tmp['Y_rmse_normal_aux'].std(),
                       'Y_rmse_normal_final_mu': tmp['Y_rmse_normal_final'].mean(),
                       'Y_rmse_normal_final_sigma': tmp['Y_rmse_normal_final'].std(),
                       'Y_rmse_extreme_aux_mu': tmp['Y_rmse_extreme_aux'].mean(),
                       'Y_rmse_extreme_aux_sigma': tmp['Y_rmse_extreme_aux'].std(),
                       'Y_rmse_extreme_final_mu': tmp['Y_rmse_extreme_final'].mean(),
                       'Y_rmse_extreme_final_sigma': tmp['Y_rmse_extreme_final'].std()
                       }
    
        
    print('********************************')
    print('********************************')
    print(site)
    print('Y_rmse_final_mu:  ',tmp['Y_rmse_final'].mean())
    print('Y_rmse_aux_mu:  ',tmp['Y_rmse_aux'].mean())
    print('Y_rmse_extreme_final_mu:  ',tmp['Y_rmse_extreme_final'].mean())
    print('Y_rmse_extreme_aux_mu:  ',tmp['Y_rmse_extreme_aux'].mean())
    print('Y_rmse_normal_final_mu:  ',tmp['Y_rmse_normal_final'].mean())
    print('Y_rmse_normal_aux_mu:  ',tmp['Y_rmse_normal_aux'].mean())
    print('********************************')
    print('********************************')
    print('')
    print('')

Y_rmse_mean_df = pd.DataFrame(Y_rmse_mean)
Y_rmse_mean_df.to_csv('BP_testsites_Yrmse_out'+'_'+loss_func+flag+'.csv', index=True, header=True)
        
    
print(loss_func+flag)