# -*- coding: utf-8 -*-
"""
Apply trained MULTT model to each site for TIME_P interval

@author: Kathy_Breen
"""

from keras.layers import Input, Dense, Concatenate, Dropout, BatchNormalization, CuDNNLSTM
from keras.models import Model
from keras.callbacks import EarlyStopping, ReduceLROnPlateau, CSVLogger, ModelCheckpoint
from keras.initializers import Ones
from keras.optimizers import Adam, Nadam, RMSprop
from sklearn.metrics import mean_squared_error
import pickle as pkl
import h5py
import numpy as np
import pandas as pd
from matplotlib import rcParams  # next 3 lines set font family for plotting
rcParams['font.family'] = 'serif'
rcParams['font.sans-serif'] = ['TImes New Roman']
import matplotlib.pyplot as plt    
# increase font size
plt.rcParams.update({'font.size': 22})
import os
#import time
import datetime
from random import shuffle
from scipy import stats

print('DEFINE WORKING DIRECTORY PATH AS THE VARIABLE "main"')
main = ''
os.chdir(main)

#%% SETTINGS FOR REPRODUCIBLE RESULTS DURING DEVELOPMENT
'''
#import numpy as np
import tensorflow as tf
import random as rn

# The below is necessary in Python 3.2.3 onwards to
# have reproducible behavior for certain hash-based operations.
# See these references for further details:
# https://docs.python.org/3.4/using/cmdline.html#envvar-PYTHONHASHSEED
# https://github.com/keras-team/keras/issues/2280#issuecomment-306959926

#import os
os.environ['PYTHONHASHSEED'] = '0'

# The below is necessary for starting Numpy generated random numbers
# in a well-defined initial state.

np.random.seed(42)

# The below is necessary for starting core Python generated random numbers
# in a well-defined state.

rn.seed(12345)

# Force TensorFlow to use single thread.
# Multiple threads are a potential source of
# non-reproducible results.
# For further details, see: https://stackoverflow.com/questions/42022950/which-seeds-have-to-be-set-where-to-realize-100-reproducibility-of-training-res

session_conf = tf.compat.v1.ConfigProto(intra_op_parallelism_threads=1, inter_op_parallelism_threads=1)

from keras import backend as K

# The below tf.set_random_seed() will make random number generation
# in the TensorFlow backend have a well-defined initial state.
# For further details, see: https://www.tensorflow.org/api_docs/python/tf/set_random_seed

#tf.set_random_seed(1234)
tf.compat.v1.set_random_seed(1234)

sess = tf.compat.v1.Session(graph=tf.compat.v1.get_default_graph(), config=session_conf)
K.set_session(sess)
'''
#%% WORK

import tensorflow as tf
from keras import backend as K

# define custom loss function

def sameSign(x,y):
    # returns 0 if x and y are of opposite sign, and 1 if the signs are the same
    xsign = tf.sign(x)
    ysign = tf.sign(y)
    
    return tf.to_float(tf.equal(xsign,ysign))
    

def pblf(ET_DATA,ET_SHIFT):
    
    # ET_DATA is an array the same size as y_true with daily ET values temporally sync all samples in Y
    # Y_SHIFT is an array the same size as y_true where y_true[i,:] == Y_SHIFT[i-1,:]
    # ET_SHIFT is an array the same size as ET_DATA where ET_DATA[i,:] == ET_SHIFT[i-1,:]
    
    def loss(y_true,y_pred):
    
        Y_SHIFT = tf.roll(y_true,1,axis=0)

        y_pred_diff = tf.subtract(y_pred,Y_SHIFT)
        et_diff = tf.subtract(ET_DATA,ET_SHIFT)
        
        # test if soil moisture decreased when et increased and v.v. - penalize loss function if physical relationship does not hold, i.e. the signs are the same
        delta_sign = sameSign(y_pred_diff,et_diff)
        return K.mean(K.square(y_pred-y_true) + 0.1*delta_sign*K.sqrt(K.square(et_diff) + K.square(y_pred_diff)),axis=-1)
    
    return loss

def LSTM_PGNN_addtnl_params(XD_train,XS_train,Yet_train,Yet_shift_train,Ysmap_train,XD_test,XS_test,Yet_test,Yet_shift_test,Ysmap_test,params):

    main_input = Input(shape=(XD_Ntimesteps,XD_Nfeatures),
                   dtype='float',
                   name='Dynamic_Input'
                   )
    x = BatchNormalization()(main_input)
       
    if params['Nlyr_LSTM']>2:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
        for i in range(1,params['Nlyr_LSTM']-1):
            lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
        
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
    elif params['Nlyr_LSTM'] == 2:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = True,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(lstm_out) 
    else:
        lstm_out = CuDNNLSTM(params['Nnodes_LSTM'], 
                            return_sequences = False,
                            unit_forget_bias=True,  
                            kernel_initializer=Ones(),  
                            recurrent_initializer=Ones()
                            )(x) 
            
    auxiliary_output = Dense(Ysmap_Nfeatures, name='aux_output', activation=params['MLP_activation'])(lstm_out)
    auxiliary_input = Input(shape=(XS_Nfeatures,)) 
    x = Concatenate(axis=-1)([lstm_out, auxiliary_input]) 
    
    x = Dense(params['Nnodes_MLP'], activation=params['MLP_activation'])(x)
    x = BatchNormalization()(x)
    Dropout(params['do'])(x)
        
    if params['Nlyr_MLP'] > 1:
        for i in range(1,params['Nlyr_MLP']):
            x = Dense(params['Nnodes_MLP'], activation=params['MLP_activation'])(x)
            x = BatchNormalization()(x)
            Dropout(params['do'])(x)
    
    final_output = Dense(Ysmap_Nfeatures, name='final_output', activation=params['MLP_activation'])(x)
    
    # create the placeholder tensors for the PBLF function
    y_et = Input(shape=(Yet_train.shape[1],), name='y_et') 
    et_shift = Input(shape=(Yet_train.shape[1],), name='et_shift') 
    
    # define model
    model = Model(inputs=[main_input, auxiliary_input, y_et, et_shift], 
                  outputs=[final_output, auxiliary_output]
                  )
    
#    adagrad = Adagrad(clipvalue=0.01)  # employ gradient clipping to avoid nans in loss fxn
    opt = params['optimizer']
    if params['TF_pblf'] ==1:
        model.compile(loss=pblf(ET_DATA=y_et,ET_SHIFT=et_shift),
                      optimizer=opt,
                      metrics=['mse'])
    else:
        model.compile(loss='mean_squared_error',
                      optimizer=opt,
                      metrics=['mse'])
    
    reduce_lr = ReduceLROnPlateau(monitor=params['monitor'],
                                  factor=params[ 'lr_factor'],  
                                  patience=params['patience_es']-5,
                                  verbose=1)
    
    early_stop = EarlyStopping(monitor=params['monitor'], 
                               min_delta=params['min_delta'], 
                               patience=params['patience_es'], 
                               verbose=1)  
    
    #print('Train the model............')
    if params['TFlr'] == 1:
        history = model.fit([XD_train,XS_train,Yet_train,Yet_shift_train],[Ysmap_train,Ysmap_train],
                epochs=epochs,
                batch_size=params['batch_size'],
                validation_data=([XD_test,XS_test,Yet_test,Yet_shift_test],[Ysmap_test,Ysmap_test]),
                callbacks=[reduce_lr],
#                callbacks=[reduce_lr,early_stop],
                verbose=2
                        )
    else:
        history = model.fit([XD_train,XS_train,Yet_train,Yet_shift_train,],[Ysmap_train,Ysmap_train],
                epochs=epochs,
                batch_size=params['batch_size'],
                validation_data=([XD_test,XS_test,Yet_test,Yet_shift_test],[Ysmap_test,Ysmap_test]),
#                callbacks=[early_stop],
                verbose=2
                )
    
    return history, model
    

#%%
    
#print('TRAIN/TEST NO FLAG')

loss_func = 'MSE'
#loss_func = 'PBLF'
flag = ''
#flag = '_noflag'


if loss_func == 'MSE':
    params = {
    'Nnodes_LSTM': 5,
    'Nnodes_MLP': 100,
    'Nlyr_LSTM': 1,
    'Nlyr_MLP': 10,
    'batch_size': 100,
    'MLP_activation': 'relu',
    'optimizer': Nadam(clipvalue=0.01),
    'do': 0.3,
    'min_delta': 0,  # min delta for early stop
    'TFlr': 0,  # true or false to use learning rate reduction (some optimizers do this anyway)
    'patience_es': 100,  # patience for early stopping
    'lr_factor': 0.8,  # factor for lr reduction
    'monitor': 'val_final_output_loss',  # quantity to monitor for callbacks
    'TF_pblf': 0  # use pblf or mse
    }
else:    
    params = {
    'Nnodes_LSTM': 5,
    'Nnodes_MLP': 100,
    'Nlyr_LSTM': 1,
    'Nlyr_MLP': 10,
    'batch_size': 75,
    'MLP_activation': 'elu',
    'optimizer': Nadam(clipvalue=0.01),
    'do': 0.2,
    'min_delta': 0.0001,  # min delta for early stop
    'TFlr': 1,  # true or false to use learning rate reduction (some optimizers do this anyway)
    'patience_es': 20,  # patience for early stopping
    'lr_factor': 0.5,  # factor for lr reduction
    'monitor': 'val_final_output_loss',  # quantity to monitor for callbacks
    'TF_pblf': 1  # use pblf or mse
    }

#%%

with h5py.File('X_D_LSTM'+flag+'.hdf5','r') as f:
    XD_train = np.array(f["XD_train"])
    XD_test = np.array(f["XD_test"])
        
with h5py.File('X_S_LSTM.hdf5','r') as f:
    XS_train = np.array(f["XS_train"])
    XS_test = np.array(f["XS_test"])
        
with h5py.File('Y_LSTM.hdf5','r') as f:
    Ysmap_train = np.array(f["Y_train"])
    Ysmap_test = np.array(f["Y_test"])
    
with h5py.File('Y_ET_LSTM.hdf5','r') as f:
    Yet_train = np.array(f["Yet_train"])
    Yet_test = np.array(f["Yet_test"])
    Yet_shift_train = np.array(f["Yet_shift_train"])
    Yet_shift_test = np.array(f["Yet_shift_test"])
        
with h5py.File('Test0'+flag+'.hdf5','r') as f:
    XD_test0 = np.array(f['XD_test0'])
    XS_test0 = np.array(f['XS_test0'])
    Ysmap_test0 = np.array(f['Y_test0'])
    Yet_test0 = np.array(f['Yet_test0'])
    Yet_shift_test0 = np.array(f["Yet_shift_test0"])

# the model is trained on 1096 timesteps - 01/01/2016-12/31/2018. the input arrays have data from 01/01/16-12/31/19 the output arrays all have data from 3/31/15-12/31/19 - limit the data I/O to the time period for training. To predict "future" SM, use the last 1096 timesteps of each array [-1096:]

XD_test0 = XD_test0[:,-1096:,:]
Ysmap_test0 = Ysmap_test0[:,-1096:]
Yet_test0 = Yet_test0[:,-1096:]
Yet_shift_test0 = Yet_shift_test0[:,-1096:]




#%%
# load saved model...for some reason I can't get the model.evaluate function to work, and model.predict only works with the dynamic and static inputs...just FYI the results are still solid
from keras.models import load_model

y_et = Input(shape=(Yet_test0.shape[1],), name='y_et') 
et_shift = Input(shape=(Yet_test0.shape[1],), name='et_shift') 

#del model
if loss_func == 'PBLF':
    from keras.utils import CustomObjectScope
    with CustomObjectScope({'loss': pblf(ET_DATA=y_et,ET_SHIFT=et_shift)}):
        model = load_model('All4sites_model'+'_'+loss_func+flag+'.h5')
else:
    model = load_model('All4sites_model'+'_'+loss_func+flag+'.h5')

histdict = pkl.load(open('All4sites_histdict'+'_'+loss_func+flag+'.pkl',"rb"))

#%% test and predict

os.chdir(main)

# read in test arrays not used during training/validation for all sites       
with h5py.File('ESCL/Test0_ESCL'+flag+'.hdf5','r') as f:
    XD_ESCL_test0 = np.array(f['XD_test0'])
    XS_ESCL_test0 = np.array(f['XS_test0'])
    Ysmap_ESCL_test0 = np.array(f['Y_test0'])
    Yet_ESCL_test0 = np.array(f['Yet_test0'])
    Yet_ESCL_shift_test0 = np.array(f['Yet_shift_test0'])
      
with h5py.File('WASH/Test0_WASH'+flag+'.hdf5','r') as f:
    XD_WASH_test0 = np.array(f['XD_test0'])
    XS_WASH_test0 = np.array(f['XS_test0'])
    Ysmap_WASH_test0 = np.array(f['Y_test0'])
    Yet_WASH_test0 = np.array(f['Yet_test0'])
    Yet_WASH_shift_test0 = np.array(f['Yet_shift_test0'])
      
with h5py.File('YAZO/Test0_YAZO'+flag+'.hdf5','r') as f:
    XD_YAZO_test0 = np.array(f['XD_test0'])
    XS_YAZO_test0 = np.array(f['XS_test0'])
    Ysmap_YAZO_test0 = np.array(f['Y_test0'])
    Yet_YAZO_test0 = np.array(f['Yet_test0'])
    Yet_YAZO_shift_test0 = np.array(f['Yet_shift_test0'])
    
XD_ESCL_test0 = XD_ESCL_test0[:,-1096:]
Ysmap_ESCL_test0 = Ysmap_ESCL_test0[:,-1096:]
Yet_ESCL_test0 = Yet_ESCL_test0[:,-1096:]
Yet_ESCL_shift_test0 = Yet_ESCL_shift_test0[:,-1096:]

XD_WASH_test0 = XD_WASH_test0[:,-1096:,:]   
Ysmap_WASH_test0 = Ysmap_WASH_test0[:,-1096:]
Yet_WASH_test0 = Yet_WASH_test0[:,-1096:]
Yet_WASH_shift_test0 = Yet_WASH_shift_test0[:,-1096:]

XD_YAZO_test0 = XD_YAZO_test0[:,-1096:,:]    
Ysmap_YAZO_test0 = Ysmap_YAZO_test0[:,-1096:]
Yet_YAZO_test0 = Yet_YAZO_test0[:,-1096:]
Yet_YAZO_shift_test0 = Yet_YAZO_shift_test0[:,-1096:]

#%%

# use for loaded model
predict = model.predict([XD_test0,XS_test0],batch_size=params['batch_size'],verbose=1)
predict_ESCL = model.predict([XD_ESCL_test0,XS_ESCL_test0],batch_size=params['batch_size'],verbose=1)
predict_WASH = model.predict([XD_WASH_test0,XS_WASH_test0],batch_size=params['batch_size'],verbose=1)
predict_YAZO = model.predict([XD_YAZO_test0,XS_YAZO_test0],batch_size=params['batch_size'],verbose=1)

Y_rmse_out = {}
timeseries_out = {}
stats_out = {}
predictions = [predict,predict_YAZO,predict_WASH,predict_ESCL]
Ysmap_test0_lst = [Ysmap_test0,Ysmap_YAZO_test0,Ysmap_WASH_test0,Ysmap_ESCL_test0]
site_ids = ['ELKR','YAZO','WASH','ESCL']

for ip,predict in enumerate(predictions):
    
    Y_rmse_out[site_ids[ip]] = {}
    
    Ysmap_test0_ip = Ysmap_test0_lst[ip]
    Yhat_smap_final = predict[0]  # final output predictions to compare with Ysmap
    Yhat_smap_aux = predict[1]  # auxilliary output predictions to compare with Ysmap
    
    # perform predictions for each of the test sites and write the results to file
    for test_site_idx in range(XD_test0.shape[0]):
        Y_rmse_final = np.sqrt(mean_squared_error(Ysmap_test0_ip[test_site_idx],Yhat_smap_final[test_site_idx]))
        Y_rmse_aux = np.sqrt(mean_squared_error(Ysmap_test0_ip[test_site_idx],Yhat_smap_aux[test_site_idx]))
        
        # calculate moving average/std dev fr SM same way it was done for input data
        window_size = 90  # calc moving avg using a window of 10 days
        stride = int(window_size*.5)  # overlap windows by 50%
        ts = Ysmap_test0_ip[test_site_idx,:]
        ts_mu = np.zeros(ts.shape)
        ts_sigma = np.zeros(ts.shape) 
        window_beg_idx = list(range(0, len(ts), stride))
        ts_ma = [np.mean(ts[i:i+window_size]) for i in window_beg_idx if i+window_size <= len(ts)]
        ts_std = [np.std(ts[i:i+window_size]) for i in window_beg_idx if i+window_size <= len(ts)]
        for i,idx in enumerate(window_beg_idx):
        #            print(i,i+window_size)
            if idx+window_size <= len(ts):
                ts_mu[idx:idx+window_size] = ts_ma[i]
                ts_sigma[idx:idx+window_size] = ts_std[i]
            else:
                ts_mu[idx:len(ts)] = ts_ma[len(ts_ma)-1]
                ts_sigma[idx:len(ts)] = ts_std[len(ts_std)-1]
                
        # predict RMSE for prediction of extreme values and "normal" values separately
        plus_sigma_idx = np.where(Ysmap_test0_ip[test_site_idx,:] >= ts_mu + ts_sigma)
        minus_sigma_idx = np.where(Ysmap_test0_ip[test_site_idx,:] <= ts_mu - ts_sigma)
        normal_idx = np.where((Ysmap_test0_ip[test_site_idx,:] > ts_mu - ts_sigma) & (Ysmap_test0_ip[test_site_idx,:] < ts_mu + ts_sigma))
        np.where(Ysmap_test0_ip[test_site_idx,:] < ts_mu + ts_sigma)[0]
        a = np.concatenate([Ysmap_test0_ip[test_site_idx,plus_sigma_idx],Ysmap_test0_ip[test_site_idx,minus_sigma_idx]],axis=1)
        b = np.concatenate([Yhat_smap_final[test_site_idx,plus_sigma_idx],Yhat_smap_final[test_site_idx,minus_sigma_idx]],axis=1)
        Y_rmse_extreme_final = np.sqrt(mean_squared_error(a,b))
        b = np.concatenate([Yhat_smap_aux[test_site_idx,plus_sigma_idx],Yhat_smap_aux[test_site_idx,minus_sigma_idx]],axis=1)
        Y_rmse_extreme_aux = np.sqrt(mean_squared_error(a,b))
        a = Ysmap_test0_ip[test_site_idx,normal_idx]
        b = Yhat_smap_final[test_site_idx,normal_idx]
        Y_rmse_normal_final = np.sqrt(mean_squared_error(a,b))
        b = Yhat_smap_aux[test_site_idx,normal_idx]
        Y_rmse_normal_aux = np.sqrt(mean_squared_error(a,b))
    
        Y_rmse_out[site_ids[ip]][test_site_idx] = {
                                   'Y_rmse_aux': Y_rmse_aux,
                                   'Y_rmse_final': Y_rmse_final,
                                   'Y_rmse_normal_aux': Y_rmse_normal_aux,
                                   'Y_rmse_normal_final': Y_rmse_normal_final,
                                   'Y_rmse_extreme_aux': Y_rmse_extreme_aux,
                                   'Y_rmse_extreme_final': Y_rmse_extreme_final
                                   }
    
    # PLOT TIMESERIES OUTPUT FROM AUX LAYER
    base = datetime.datetime.strptime("2017-01-01", "%Y-%m-%d")
    date_lst = [base + datetime.timedelta(days=x) for x in range(1096)]
    xticks = pd.date_range(date_lst[0],date_lst[len(date_lst)-1],freq='QS').tolist()
    xticklabels = []
    for tick in xticks:
        if tick.month == 1:
            label = 'Jan -' + str(tick.year)
            xticklabels.append(label)
        elif tick.month == 7:
            label = 'Jul -' + str(tick.year)
            xticklabels.append(label)
        else:
            xticklabels.append(' ')
    
    # store timeseries and stats
    stats_out[site_ids[ip]] = {}
    
    timeseries_out['DATE'] = []
    for date in date_lst:
        timeseries_out['DATE'].append(date.strftime('%m/%d/%Y'))
    timeseries_out['Y_SMAP_'+site_ids[ip]] = Ysmap_test0_ip.mean(axis=0)
    timeseries_out['Y_AUX_'+site_ids[ip]] = Yhat_smap_aux.mean(axis=0)
    timeseries_out['Y_FINAL_'+site_ids[ip]] = Yhat_smap_final.mean(axis=0)
    
    aux_slope, aux_intercept, aux_r_value, aux_p_value, aux_std_err = stats.linregress(Ysmap_test0_ip.mean(axis=0),Yhat_smap_aux.mean(axis=0))
    final_slope, final_intercept, final_r_value, final_p_value, final_std_err = stats.linregress(Ysmap_test0_ip.mean(axis=0),Yhat_smap_final.mean(axis=0))
    stats_out[site_ids[ip]]['R2_AUX'] = aux_r_value**2
    stats_out[site_ids[ip]]['R2_FINAL'] = final_r_value**2
    
    # PLOT TIMESERIES OUTPUT FROM FINAL LAYER
    fig = plt.figure(figsize=(8,6))
    ax1 = fig.add_subplot(111)
    yhat_aux = ax1.plot(date_lst,Yhat_smap_aux.mean(axis=0),'b-',label=r'$\hat{Y}_{\mathrm{SMAP,aux}}$',linewidth=6)
    yhat_final = ax1.plot(date_lst,Yhat_smap_final.mean(axis=0),'c-',label=r'$\hat{Y}_{\mathrm{SMAP,final}}$',linewidth=6)
    y = ax1.plot(date_lst,Ysmap_test0_ip.mean(axis=0),'k--', label=r'$Y_{\mathrm{SMAP}}$', linewidth=3)
    ax1.set_xlim(left=date_lst[0],right=date_lst[len(date_lst)-1])
    ax1.set_ylim((0.0,0.5))
    ax1.set_xticks(xticks)
    ax1.set_xticklabels(xticklabels,rotation=30)
    ax1.set_xlabel('Date')
    ax1.set_ylabel(r'Soil moisture ($\mathrm{cm}^3 \mathrm{cm}^{-3})$')

    plt.tight_layout()
    plt.savefig('All4sites_'+site_ids[ip]+'_y_yhat'+'_'+loss_func+'2017-2019'+flag+'.png', format='png')
    plt.show()
    
    # PLOT Y V YHAT AUX
    fig = plt.figure(figsize=(8,6))
    ax = fig.add_subplot(111)
    one2one = ax.plot([0.0,0.05,0.1,0.2,0.3,0.4,0.5],[0.0,0.05,0.1,0.2,0.3,0.4,0.5],'k-')
    yvyhat_aux = ax.plot(Ysmap_test0_ip.mean(axis=0),Yhat_smap_aux.mean(axis=0),'b*',label=r'$\hat{Y}_{\mathrm{SMAP,aux}}$',markersize=10)
    yvyhat_final = ax.plot(Ysmap_test0_ip.mean(axis=0),Yhat_smap_final.mean(axis=0),'c*',label=r'$\hat{Y}_{SMAP,final}$',markersize=10)
    ax.set_ylabel(r'$\hat{Y}_{\mathrm{SMAP}}$')
    ax.set_xlabel(r'$Y_{\mathrm{SMAP}}$')
    ax.set_ylim((0.0,0.5))
    ax.set_xlim((0.0,0.5))
    plt.tight_layout()
#    plt.savefig('All4sites_'+site_ids[ip]+'_y_v_yhat'+'_'+loss_func+'2017-2019'+flag+'.png')
    plt.show()
    
    plt.close('all')
    
# write output to file
Y_rmse_out_df = pd.DataFrame.from_dict({(i,j): Y_rmse_out[i][j] 
                           for i in Y_rmse_out.keys() 
                           for j in Y_rmse_out[i].keys()},
                       orient='index')

Y_rmse_out_df.to_csv('RAW_All4_Yrmse_out'+'_'+loss_func+'2017-2019'+flag+'.csv', index=True, header=True)

Y_rmse_mean = {}
for site in site_ids:
    tmp = Y_rmse_out_df[[site in s for s in Y_rmse_out_df.index]]
    Y_rmse_mean[site] = {
                       'Y_rmse_aux_mu': tmp['Y_rmse_aux'].mean(),
                       'Y_rmse_aux_sigma': tmp['Y_rmse_aux'].std(),
                       'Y_rmse_final_mu': tmp['Y_rmse_final'].mean(),
                       'Y_rmse_final_sigma': tmp['Y_rmse_final'].std(),
                       'Y_rmse_normal_aux_mu': tmp['Y_rmse_normal_aux'].mean(),
                       'Y_rmse_normal_aux_sigma': tmp['Y_rmse_normal_aux'].std(),
                       'Y_rmse_normal_final_mu': tmp['Y_rmse_normal_final'].mean(),
                       'Y_rmse_normal_final_sigma': tmp['Y_rmse_normal_final'].std(),
                       'Y_rmse_extreme_aux_mu': tmp['Y_rmse_extreme_aux'].mean(),
                       'Y_rmse_extreme_aux_sigma': tmp['Y_rmse_extreme_aux'].std(),
                       'Y_rmse_extreme_final_mu': tmp['Y_rmse_extreme_final'].mean(),
                       'Y_rmse_extreme_final_sigma': tmp['Y_rmse_extreme_final'].std()
                       }
    
        
    print('********************************')
    print('********************************')
    print(site)
    print('Y_rmse_final_mu:  ',tmp['Y_rmse_final'].mean())
    print('Y_rmse_aux_mu:  ',tmp['Y_rmse_aux'].mean())
    print('Y_rmse_extreme_final_mu:  ',tmp['Y_rmse_extreme_final'].mean())
    print('Y_rmse_extreme_aux_mu:  ',tmp['Y_rmse_extreme_aux'].mean())
    print('Y_rmse_normal_final_mu:  ',tmp['Y_rmse_normal_final'].mean())
    print('Y_rmse_normal_aux_mu:  ',tmp['Y_rmse_normal_aux'].mean())
    print('********************************')
    print('********************************')
    print('')
    print('')

Y_rmse_mean_df = pd.DataFrame(Y_rmse_mean)
Y_rmse_mean_df.to_csv('All4_Yrmse_out'+'_'+loss_func+'2017-2019'+flag+'.csv', index=True, header=True)

timeseries_df = pd.DataFrame(timeseries_out)
stats_df = pd.DataFrame(stats_out)

timeseries_df.to_csv('MULTT_TIME_P_'+loss_func+flag+'_TIMESERIES.csv',header=True,index=False)
stats_df.to_csv('MULTT_TIME_P_'+loss_func+flag+'_STATS.csv',header=True,index=True)































































